/*
 * Copyright (C) 1994-2016 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *  
 * This file is part of the PBS Professional ("PBS Pro") software.
 * 
 * Open Source License Information:
 *  
 * PBS Pro is free software. You can redistribute it and/or modify it under the
 * terms of the GNU Affero General Public License as published by the Free 
 * Software Foundation, either version 3 of the License, or (at your option) any 
 * later version.
 *  
 * PBS Pro is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License along 
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Commercial License Information: 
 * 
 * The PBS Pro software is licensed under the terms of the GNU Affero General 
 * Public License agreement ("AGPL"), except where a separate commercial license 
 * agreement for PBS Pro version 14 or later has been executed in writing with Altair.
 *  
 * Altair’s dual-license business model allows companies, individuals, and 
 * organizations to create proprietary derivative works of PBS Pro and distribute 
 * them - whether embedded or bundled with other software - under a commercial 
 * license agreement.
 * 
 * Use of Altair’s trademarks, including but not limited to "PBS™", 
 * "PBS Professional®", and "PBS Pro™" and Altair’s logos is subject to Altair's 
 * trademark licensing policies.
 *
 */


/**
 * @file    db_postgres_job.c
 *
 * @brief
 *      Implementation of the job data access functions for postgres
 */

#include <pbs_config.h>   /* the master config generated by configure */
#include "pbs_db.h"
#include "db_postgres.h"

/**
 * @brief
 *	Prepare all the job related sqls. Typically called after connect
 *	and before any other sql exeuction
 *
 * @param[in]	conn - Database connection handle
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_prepare_job_sqls(pbs_db_conn_t *conn)
{
	sprintf(conn->conn_sql, "insert into pbs.job ("
		"ji_jobid,"
		"ji_sv_name,"
		"ji_state,"
		"ji_substate,"
		"ji_svrflags,"
		"ji_numattr,"
		"ji_ordering,"
		"ji_priority,"
		"ji_stime,"
		"ji_endtBdry,"
		"ji_queue,"
		"ji_destin,"
		"ji_un_type,"
		"ji_momaddr,"
		"ji_momport,"
		"ji_exitstat,"
		"ji_quetime,"
		"ji_rteretry,"
		"ji_fromsock,"
		"ji_fromaddr,"
		"ji_4jid,"
		"ji_4ash,"
		"ji_credtype,"
		"ji_qrank,"
		"ji_savetm,"
		"ji_creattm"
		") "
		"values ($1, $2, $3, $4, $5, $6, $7, $8, $9, "
		"$10, $11, $12, $13,"
		"$14, $15, $16, $17, $18, $19, $20, $21, $22, $23,"
		"$24, localtimestamp, localtimestamp)");

	if (pg_prepare_stmt(conn, STMT_INSERT_JOB, conn->conn_sql, 24) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.job set "
		"ji_sv_name = $2,"
		"ji_state = $3,"
		"ji_substate = $4,"
		"ji_svrflags = $5,"
		"ji_numattr = $6,"
		"ji_ordering = $7,"
		"ji_priority = $8,"
		"ji_stime = $9,"
		"ji_endtBdry = $10,"
		"ji_queue  = $11,"
		"ji_destin = $12,"
		"ji_un_type = $13,"
		"ji_momaddr = $14,"
		"ji_momport  = $15,"
		"ji_exitstat = $16,"
		"ji_quetime = $17,"
		"ji_rteretry = $18,"
		"ji_fromsock = $19,"
		"ji_fromaddr = $20,"
		"ji_4jid = $21,"
		"ji_4ash = $22,"
		"ji_credtype = $23,"
		"ji_qrank = $24,"
		"ji_savetm = localtimestamp"
		" where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_UPDATE_JOB, conn->conn_sql, 24) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"ji_jobid,"
		"ji_sv_name,"
		"ji_state,"
		"ji_substate,"
		"ji_svrflags,"
		"ji_numattr,"
		"ji_ordering,"
		"ji_priority,"
		"ji_stime,"
		"ji_endtBdry,"
		"ji_queue,"
		"ji_destin,"
		"ji_un_type,"
		"ji_momaddr,"
		"ji_momport,"
		"ji_exitstat,"
		"ji_quetime,"
		"ji_rteretry,"
		"ji_fromsock,"
		"ji_fromaddr,"
		"ji_4jid,"
		"ji_4ash,"
		"ji_credtype,"
		"extract(epoch from ji_savetm)::bigint as ji_savetm, "
		"extract(epoch from ji_creattm)::bigint as ji_creattm "
		"from pbs.job where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_JOB, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "insert into pbs.job_attr "
		"(ji_jobid, attr_name, attr_resource, attr_value, attr_flags) "
		"values ($1, $2, $3, $4, $5)");
	if (pg_prepare_stmt(conn, STMT_INSERT_JOBATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.job_attr set "
		"attr_resource = $3,"
		"attr_value = $4,"
		"attr_flags = $5"
		" where ji_jobid = $1 and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_UPDATE_JOBATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.job_attr set "
		"attr_value = $4,"
		"attr_flags = $5"
		" where ji_jobid = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_UPDATE_JOBATTR_RESC, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"attr_name, attr_resource, attr_value, attr_flags "
		"from pbs.job_attr "
		"where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_JOBATTR, conn->conn_sql, 1) != 0)
		return -1;

	/*
	 * Use the sql encode function to encode the $2 parameter. Encode using
	 * 'escape' mode. Encode considers $2 as a bytea and returns a escaped
	 * string using 'escape' syntax. Refer to the following postgres link
	 * for details:
	 * http://www.postgresql.org/docs/8.3/static/functions-string.html
	 */
	sprintf(conn->conn_sql, "insert into "
		"pbs.job_scr (ji_jobid, script) "
		"values "
		"($1, encode($2, 'escape'))");
	if (pg_prepare_stmt(conn, STMT_INSERT_JOBSCR, conn->conn_sql, 2) != 0)
		return -1;

	/*
	 * Use the sql decode function to decode the script parameter. Decode
	 * using 'escape' mode. Decode considers script as encoded TEXT and
	 * decodes it using 'escape' syntax, returning a bytea. The :: is used
	 * to "typecast" the output to a bytea.
	 * Refer to the following postgres link for details:
	 * http://www.postgresql.org/docs/8.3/static/functions-string.html
	 */
	sprintf(conn->conn_sql, "select decode(script, 'escape')::bytea as script "
		"from pbs.job_scr "
		"where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_JOBSCR, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"ji_jobid,"
		"ji_sv_name,"
		"ji_state,"
		"ji_substate,"
		"ji_svrflags,"
		"ji_numattr,"
		"ji_ordering,"
		"ji_priority,"
		"ji_stime,"
		"ji_endtBdry,"
		"ji_queue,"
		"ji_destin,"
		"ji_un_type,"
		"ji_momaddr,"
		"ji_momport,"
		"ji_exitstat,"
		"ji_quetime,"
		"ji_rteretry,"
		"ji_fromsock,"
		"ji_fromaddr,"
		"ji_4jid,"
		"ji_4ash,"
		"ji_credtype,"
		"extract(epoch from ji_savetm)::bigint as ji_savetm, "
		"extract(epoch from ji_creattm)::bigint as ji_creattm "
		"from pbs.job order by ji_qrank");
	if (pg_prepare_stmt(conn, STMT_FINDJOBS_ORDBY_QRANK, conn->conn_sql, 0) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"ji_jobid,"
		"ji_sv_name,"
		"ji_state,"
		"ji_substate,"
		"ji_svrflags,"
		"ji_numattr,"
		"ji_ordering,"
		"ji_priority,"
		"ji_stime,"
		"ji_endtBdry,"
		"ji_queue,"
		"ji_destin,"
		"ji_un_type,"
		"ji_momaddr,"
		"ji_momport,"
		"ji_exitstat,"
		"ji_quetime,"
		"ji_rteretry,"
		"ji_fromsock,"
		"ji_fromaddr,"
		"ji_4jid,"
		"ji_4ash,"
		"ji_credtype,"
		"extract(epoch from ji_savetm)::bigint as ji_savetm, "
		"extract(epoch from ji_creattm)::bigint as ji_creattm "
		"from pbs.job where ji_queue = $1"
		" order by ji_qrank");
	if (pg_prepare_stmt(conn, STMT_FINDJOBS_BYQUE_ORDBY_QRANK,
		conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.job where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_DELETE_JOB, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.job_scr where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_DELETE_JOBSCR, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.job_attr where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_DELETE_JOBATTR_ALL, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.job_attr "
		" where ji_jobid = $1 and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_DELETE_JOBATTR, conn->conn_sql, 2) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from "
		"pbs.job_attr "
		"where "
		"ji_jobid = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_DELETE_JOBATTR_RESC, conn->conn_sql, 3) != 0)
		return -1;

	sprintf(conn->conn_sql, "insert into pbs.subjob_track "
		"("
		"ji_jobid, "
		"trk_index, "
		"trk_status, "
		"trk_error, "
		"trk_exitstat, "
		"trk_substate, "
		"trk_stgout "
		") "
		"values "
		"($1, $2, $3, $4, $5, $6, $7)");
	if (pg_prepare_stmt(conn, STMT_INSERT_SUBJOB, conn->conn_sql, 7) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.subjob_track set "
		"trk_status = $3, "
		"trk_error = $4, "
		"trk_exitstat = $5, "
		"trk_substate = $6, "
		"trk_stgout = $7 "
		"where ji_jobid = $1 and trk_index = $2");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SUBJOB, conn->conn_sql, 7) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"ji_jobid, "
		"trk_index, "
		"trk_status, "
		"trk_error, "
		"trk_exitstat, "
		"trk_substate, "
		"trk_stgout "
		"from pbs.subjob_track "
		"where ji_jobid = $1 order by trk_index");
	if (pg_prepare_stmt(conn, STMT_FIND_SUBJOB_ORDBY_INDEX,
		conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.subjob_track where ji_jobid = $1");
	if (pg_prepare_stmt(conn, STMT_DELETE_SUBJOB, conn->conn_sql, 1) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Load job data from the row into the job object
 *
 * @param[in]	res - Resultset from a earlier query
 * @param[in]	pj  - Job object to load data into
 * @param[in]	row - The current row to load within the resultset
 *
 */
static void
load_job(PGresult *res, pbs_db_job_info_t *pj, int row)
{
	strcpy(pj->ji_jobid, PQgetvalue(res, row, PQfnumber(res, "ji_jobid")));
	pj->ji_state = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_state")), NULL, 10);
	pj->ji_substate = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_substate")), NULL, 10);
	pj->ji_svrflags = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_svrflags")), NULL, 10);
	pj->ji_numattr = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_numattr")), NULL, 10);
	pj->ji_ordering = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_ordering")), NULL, 10);
	pj->ji_priority = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_priority")), NULL, 10);
	pj->ji_stime = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_stime")), NULL, 10);
	pj->ji_endtBdry = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_endtBdry")), NULL, 10);
	strcpy(pj->ji_queue, PQgetvalue(res, row, PQfnumber(res, "ji_queue")));
	strcpy(pj->ji_destin, PQgetvalue(res, row,
		PQfnumber(res, "ji_destin")));
	pj->ji_credtype = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_credtype")), NULL, 10);
	strcpy(pj->ji_4jid, PQgetvalue(res, row, PQfnumber(res, "ji_4jid")));
	strcpy(pj->ji_4ash, PQgetvalue(res, row, PQfnumber(res, "ji_4ash")));
	pj->ji_un_type = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_un_type")), NULL, 10);
	pj->ji_momaddr = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_momaddr")), NULL, 10);
	pj->ji_momport = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_momport")), NULL, 10);
	pj->ji_exitstat = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_exitstat")), NULL, 10);
	pj->ji_quetime = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_quetime")), NULL, 10);
	pj->ji_rteretry = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_rteretry")), NULL, 10);
	pj->ji_fromsock = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ji_fromsock")), NULL, 10);
	pj->ji_fromaddr = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_fromaddr")), NULL, 10);
	strcpy(pj->ji_sv_name, PQgetvalue(res, row,
		PQfnumber(res, "ji_sv_name")));
	pj->ji_savetm = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_savetm")), NULL, 10);
	pj->ji_creattm = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ji_creattm")), NULL, 10);
}

/**
 * @brief
 *	Insert job data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of job to be inserted
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_insert_job(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_job_info_t *pjob = obj->pbs_db_un.pbs_db_job;

	LOAD_STR(conn, pjob->ji_jobid, 0);
	LOAD_STR(conn, pjob->ji_sv_name, 1);
	LOAD_INTEGER(conn, pjob->ji_state, 2);
	LOAD_INTEGER(conn, pjob->ji_substate, 3);
	LOAD_INTEGER(conn, pjob->ji_svrflags, 4);
	LOAD_INTEGER(conn, pjob->ji_numattr, 5);
	LOAD_INTEGER(conn, pjob->ji_ordering, 6);
	LOAD_INTEGER(conn, pjob->ji_priority, 7);
	LOAD_BIGINT(conn, pjob->ji_stime, 8);
	LOAD_BIGINT(conn, pjob->ji_endtBdry, 9);
	LOAD_STR(conn, pjob->ji_queue, 10);
	LOAD_STR(conn, pjob->ji_destin, 11);
	LOAD_INTEGER(conn, pjob->ji_un_type, 12);
	LOAD_BIGINT(conn, pjob->ji_momaddr, 13);
	LOAD_INTEGER(conn, pjob->ji_momport, 14);
	LOAD_INTEGER(conn, pjob->ji_exitstat, 15);
	LOAD_BIGINT(conn, pjob->ji_quetime, 16);
	LOAD_BIGINT(conn, pjob->ji_rteretry, 17);
	LOAD_INTEGER(conn, pjob->ji_fromsock, 18);
	LOAD_BIGINT(conn, pjob->ji_fromaddr, 19);
	LOAD_STR(conn, pjob->ji_4jid, 20);
	LOAD_STR(conn, pjob->ji_4ash, 21);
	LOAD_INTEGER(conn, pjob->ji_credtype, 22);
	LOAD_INTEGER(conn, pjob->ji_qrank, 23);

	if (pg_db_cmd(conn, STMT_INSERT_JOB, 24) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Update job data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of job to be updated
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows updated
 *
 */
int
pg_db_update_job(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_job_info_t *pjob = obj->pbs_db_un.pbs_db_job;

	LOAD_STR(conn, pjob->ji_jobid, 0);
	LOAD_STR(conn, pjob->ji_sv_name, 1);
	LOAD_INTEGER(conn, pjob->ji_state, 2);
	LOAD_INTEGER(conn, pjob->ji_substate, 3);
	LOAD_INTEGER(conn, pjob->ji_svrflags, 4);
	LOAD_INTEGER(conn, pjob->ji_numattr, 5);
	LOAD_INTEGER(conn, pjob->ji_ordering, 6);
	LOAD_INTEGER(conn, pjob->ji_priority, 7);
	LOAD_BIGINT(conn, pjob->ji_stime, 8);
	LOAD_BIGINT(conn, pjob->ji_endtBdry, 9);
	LOAD_STR(conn, pjob->ji_queue, 10);
	LOAD_STR(conn, pjob->ji_destin, 11);
	LOAD_INTEGER(conn, pjob->ji_un_type, 12);
	LOAD_BIGINT(conn, pjob->ji_momaddr, 13);
	LOAD_INTEGER(conn, pjob->ji_momport, 14);
	LOAD_INTEGER(conn, pjob->ji_exitstat, 15);
	LOAD_BIGINT(conn, pjob->ji_quetime, 16);
	LOAD_BIGINT(conn, pjob->ji_rteretry, 17);
	LOAD_INTEGER(conn, pjob->ji_fromsock, 18);
	LOAD_BIGINT(conn, pjob->ji_fromaddr, 19);
	LOAD_STR(conn, pjob->ji_4jid, 20);
	LOAD_STR(conn, pjob->ji_4ash, 21);
	LOAD_INTEGER(conn, pjob->ji_credtype, 22);
	LOAD_INTEGER(conn, pjob->ji_qrank, 23);

	return (pg_db_cmd(conn, STMT_UPDATE_JOB, 24));
}

/**
 * @brief
 *	Load job data from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in/out]obj  - Load job information into this object where
 *			jobid = obj->pbs_db_un.pbs_db_job->ji_jobid
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows loaded
 *
 */
int
pg_db_load_job(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	PGresult *res;
	int rc;
	pbs_db_job_info_t *pj = obj->pbs_db_un.pbs_db_job;

	LOAD_STR(conn, pj->ji_jobid, 0);

	if ((rc = pg_db_query(conn, STMT_SELECT_JOB, 1, &res)) != 0)
		return rc;

	/* get the other fields */
	load_job(res, pj, 0);

	PQclear(res);
	return 0;
}

/**
 * @brief
 *	Find jobs
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state variable updated by this query
 * @param[in]	obj  - Information of job to be found
 * @param[in]	opts - Any other options (like flags, timestamp)
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows found
 *
 */
int
pg_db_find_job(pbs_db_conn_t *conn, void *st, pbs_db_obj_info_t *obj,
	pbs_db_query_options_t *opts)
{
	PGresult *res;
	pg_query_state_t *state = (pg_query_state_t *) st;
	pbs_db_job_info_t *pdjob = obj->pbs_db_un.pbs_db_job;
	int rc;
	int params;

	if (!state)
		return -1;

	if (opts != NULL && opts->flags == FIND_JOBS_BY_QUE) {
		LOAD_STR(conn, pdjob->ji_queue, 0);
		params=1;
		strcpy(conn->conn_sql, STMT_FINDJOBS_BYQUE_ORDBY_QRANK);
	} else {
		strcpy(conn->conn_sql, STMT_FINDJOBS_ORDBY_QRANK);
		params=0;
	}

	if ((rc = pg_db_query(conn, conn->conn_sql, params, &res)) != 0)
		return rc;

	state->row = 0;
	state->res = res;
	state->count = PQntuples(res);

	return 0;
}

/**
 * @brief
 *	Get the next job from the cursor
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state
 * @param[in]	obj  - Job information is loaded into this object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_next_job(pbs_db_conn_t *conn, void *st, pbs_db_obj_info_t *obj)
{
	pg_query_state_t *state = (pg_query_state_t *) st;

	load_job(state->res, obj->pbs_db_un.pbs_db_job, state->row);

	return 0;
}

/**
 * @brief
 *	Delete the job from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Job information
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows deleted
 *
 */
int
pg_db_delete_job(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_job_info_t *pj = obj->pbs_db_un.pbs_db_job;
	int rc = 0;

	LOAD_STR(conn, pj->ji_jobid, 0);

	if (pbs_db_begin_trx(conn, 0, 0) != 0)
		goto err;

	if ((rc = pg_db_cmd(conn, STMT_DELETE_JOB, 1)) == -1)
		goto err;

	if (pg_db_cmd(conn, STMT_DELETE_JOBATTR_ALL, 1) == -1)
		goto err;

	if (pg_db_cmd(conn, STMT_DELETE_JOBSCR, 1) == -1)
		goto err;

	if (pg_db_cmd(conn, STMT_DELETE_SUBJOB, 1) == -1)
		goto err;

	if (pbs_db_end_trx(conn, PBS_DB_COMMIT) != 0)
		goto err;

	return rc;
err:
	(void) pbs_db_end_trx(conn, PBS_DB_ROLLBACK);
	return -1;
}

/**
 * @brief
 *	Insert job script
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Job script object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_insert_jobscr(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_jobscr_info_t *pscr = obj->pbs_db_un.pbs_db_jobscr;

	LOAD_STR(conn, pscr->ji_jobid, 0);

	/*
	 * The script data could contain non-UTF8 characters. We therefore
	 * consider it binary and encode it into TEXT by using the "encode"
	 * sql function. The input data to load, therefore, is binary data
	 * and so we use the function "LOAD_BIN" to load the parameter to
	 * the prepared statement
	 */
	LOAD_BIN(conn, pscr->script, (pscr->script)?strlen(pscr->script):0, 1);

	if ((pg_db_cmd(conn, STMT_INSERT_JOBSCR, 2)) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	load job script
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Job script is loaded into this object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_load_jobscr(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	PGresult *res;
	pbs_db_jobscr_info_t *pscr = obj->pbs_db_un.pbs_db_jobscr;

	LOAD_STR(conn, pscr->ji_jobid, 0);

	/*
	 * The data (script) we stored was a "encoded" binary. We "decode" it
	 * back while reading, giving us "binary" data. Since we want the
	 * result data to be returned in binary, we set conn_result_format
	 * to 1 to indicate binary result. This setting is a one-time,
	 * auto-reset switch which resets to 0 (TEXT) mode after each execution
	 * of pg_db_query.
	 *
	 */
	conn->conn_result_format = 1; /* need binary result */
	if (pg_db_query(conn, STMT_SELECT_JOBSCR, 1, &res) != 0)
		return -1;

	pscr->script = PQgetvalue(res, 0, PQfnumber(res, "script"));

	return 0;
}

/**
 * @brief
 *	Insert subjob information into database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Subjob information
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_insert_subjob(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_subjob_info_t *psubjob = obj->pbs_db_un.pbs_db_subjob;

	LOAD_STR(conn, psubjob->ji_jobid, 0);
	LOAD_INTEGER(conn, psubjob->trk_index, 1);
	LOAD_INTEGER(conn, psubjob->trk_status, 2);
	LOAD_INTEGER(conn, psubjob->trk_error, 3);
	LOAD_INTEGER(conn, psubjob->trk_exitstat, 4);
	LOAD_INTEGER(conn, psubjob->trk_substate, 5);
	LOAD_INTEGER(conn, psubjob->trk_stgout, 6);

	if (pg_db_cmd(conn, STMT_INSERT_SUBJOB, 7) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Update subjob information in the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Subjob information to update
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows updated
 *
 */
int
pg_db_update_subjob(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_subjob_info_t *psubjob = obj->pbs_db_un.pbs_db_subjob;

	LOAD_STR(conn, psubjob->ji_jobid, 0);
	LOAD_INTEGER(conn, psubjob->trk_index, 1);
	LOAD_INTEGER(conn, psubjob->trk_status, 2);
	LOAD_INTEGER(conn, psubjob->trk_error, 3);
	LOAD_INTEGER(conn, psubjob->trk_exitstat, 4);
	LOAD_INTEGER(conn, psubjob->trk_substate, 5);
	LOAD_INTEGER(conn, psubjob->trk_stgout, 6);

	return (pg_db_cmd(conn, STMT_UPDATE_SUBJOB, 7));
}

/**
 * @brief
 *	Find a subjob related to a jobid
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state
 * @param[in]	obj  - Job information is loaded into this object
 * @param[in]	opts - Options (flag & timestamp)
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows found
 *
 */
int
pg_db_find_subjob(pbs_db_conn_t *conn, void *st, pbs_db_obj_info_t *obj,
	pbs_db_query_options_t *opts)
{
	PGresult *res;
	pg_query_state_t *state = (pg_query_state_t *) st;
	pbs_db_subjob_info_t *psubjob = obj->pbs_db_un.pbs_db_subjob;
	int rc;
	int params;

	if (!state)
		return -1;

	LOAD_STR(conn, psubjob->ji_jobid, 0);
	params=1;
	strcpy(conn->conn_sql, STMT_FIND_SUBJOB_ORDBY_INDEX);
	if ((rc = pg_db_query(conn, conn->conn_sql, params, &res)) != 0)
		return rc;

	state->row = 0;
	state->res = res;
	state->count = PQntuples(res);

	return 0;
}

/**
 * @brief
 *	load subjob information into object
 *
 * @param[in]	res  - resultset from an earlier query
 * @param[in]	psubjob  - Subjob object to load into
 * @param[in]	row - The row in the resultset to work on
 *
 */
static void
load_subjob(PGresult *res, pbs_db_subjob_info_t *psubjob, int row)
{
	strcpy(psubjob->ji_jobid, PQgetvalue(res, row,
		PQfnumber(res, "ji_jobid")));
	psubjob->trk_index = strtol(PQgetvalue(res, row,
		PQfnumber(res, "trk_index")), NULL, 10);
	psubjob->trk_status = strtol(PQgetvalue(res, row,
		PQfnumber(res, "trk_status")), NULL, 10);
	psubjob->trk_error = strtol(PQgetvalue(res, row,
		PQfnumber(res, "trk_error")), NULL, 10);
	psubjob->trk_exitstat = strtol(PQgetvalue(res, row,
		PQfnumber(res, "trk_exitstat")), NULL, 10);
	psubjob->trk_substate = strtol(PQgetvalue(res, row,
		PQfnumber(res, "trk_substate")), NULL, 10);
	psubjob->trk_stgout = strtol(PQgetvalue(res, row,
		PQfnumber(res, "trk_stgout")), NULL, 10);
}

/**
 * @brief
 *	Get the next subjob from the cursor
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state
 * @param[in]	obj  - SubJob information is loaded into this object
 *
 * @return      Error code
 *		(Even though this returns only 0 now, keeping it as int
 *			to support future change to return a failure)
 * @retval	 0 - Success
 *
 */
int
pg_db_next_subjob(pbs_db_conn_t *conn, void *st, pbs_db_obj_info_t *obj)
{
	pg_query_state_t *state = (pg_query_state_t *) st;
	pbs_db_subjob_info_t *psubjob = obj->pbs_db_un.pbs_db_subjob;

	load_subjob(state->res, psubjob, state->row);

	return 0;
}
