/*
 * Copyright (C) 1994-2016 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *  
 * This file is part of the PBS Professional ("PBS Pro") software.
 * 
 * Open Source License Information:
 *  
 * PBS Pro is free software. You can redistribute it and/or modify it under the
 * terms of the GNU Affero General Public License as published by the Free 
 * Software Foundation, either version 3 of the License, or (at your option) any 
 * later version.
 *  
 * PBS Pro is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License along 
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Commercial License Information: 
 * 
 * The PBS Pro software is licensed under the terms of the GNU Affero General 
 * Public License agreement ("AGPL"), except where a separate commercial license 
 * agreement for PBS Pro version 14 or later has been executed in writing with Altair.
 *  
 * Altair’s dual-license business model allows companies, individuals, and 
 * organizations to create proprietary derivative works of PBS Pro and distribute 
 * them - whether embedded or bundled with other software - under a commercial 
 * license agreement.
 * 
 * Use of Altair’s trademarks, including but not limited to "PBS™", 
 * "PBS Professional®", and "PBS Pro™" and Altair’s logos is subject to Altair's 
 * trademark licensing policies.
 *
 */


/**
 * @file    db_postgres_resv.c
 *
 * @brief
 *      Implementation of the resv data access functions for postgres
 */
#include <pbs_config.h>   /* the master config generated by configure */
#include "pbs_db.h"
#include "db_postgres.h"

/**
 * @brief
 *	Prepare all the reservation related sqls. Typically called after connect
 *	and before any other sql exeuction
 *
 * @param[in]	conn - Database connection handle
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_prepare_resv_sqls(pbs_db_conn_t *conn)
{
	sprintf(conn->conn_sql, "insert into pbs.resv ("
		"ri_resvID, "
		"ri_sv_name, "
		"ri_queue, "
		"ri_state, "
		"ri_substate, "
		"ri_type, "
		"ri_stime, "
		"ri_etime, "
		"ri_duration, "
		"ri_tactive, "
		"ri_svrflags, "
		"ri_numattr, "
		"ri_resvTag, "
		"ri_un_type, "
		"ri_fromsock, "
		"ri_fromaddr, "
		"ri_savetm, "
		"ri_creattm "
		") "
		"values "
		"($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, "
		"$14, $15, $16, localtimestamp, localtimestamp)");
	if (pg_prepare_stmt(conn, STMT_INSERT_RESV, conn->conn_sql, 16) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.resv set "
		"ri_sv_name = $2, "
		"ri_queue = $3, "
		"ri_state = $4, "
		"ri_substate = $5, "
		"ri_type = $6, "
		"ri_stime = $7, "
		"ri_etime = $8, "
		"ri_duration = $9, "
		"ri_tactive = $10, "
		"ri_svrflags = $11, "
		"ri_numattr = $12, "
		"ri_resvTag = $13, "
		"ri_un_type = $14, "
		"ri_fromsock = $15, "
		"ri_fromaddr = $16, "
		"ri_savetm = localtimestamp "
		"where ri_resvID = $1");
	if (pg_prepare_stmt(conn, STMT_UPDATE_RESV, conn->conn_sql, 16) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"ri_resvID, "
		"ri_sv_name, "
		"ri_queue, "
		"ri_state, "
		"ri_substate, "
		"ri_type, "
		"ri_stime, "
		"ri_etime, "
		"ri_duration, "
		"ri_tactive, "
		"ri_svrflags, "
		"ri_numattr, "
		"ri_resvTag, "
		"ri_un_type, "
		"ri_fromsock, "
		"ri_fromaddr, "
		"extract(epoch from ri_savetm)::bigint as ri_savetm, "
		"extract(epoch from ri_creattm)::bigint as ri_creattm "
		"from pbs.resv where ri_resvid = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_RESV, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "insert into "
		"pbs.resv_attr "
		"(ri_resvid, "
		"attr_name, "
		"attr_resource, "
		"attr_value, "
		"attr_flags) "
		"values "
		"($1, $2, $3, $4, $5)");
	if (pg_prepare_stmt(conn, STMT_INSERT_RESVATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.resv_attr set "
		"attr_resource = $3, "
		"attr_value = $4, "
		"attr_flags = $5 "
		"where ri_resvid = $1 "
		"and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_UPDATE_RESVATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.resv_attr set "
		"attr_value = $4, "
		"attr_flags = $5 "
		"where ri_resvid = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_UPDATE_RESVATTR_RESC, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.resv_attr "
		"where ri_resvid = $1 "
		"and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_DELETE_RESVATTR, conn->conn_sql, 2) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.resv_attr "
		"where ri_resvid = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_DELETE_RESVATTR_RESC, conn->conn_sql, 3) != 0)
		return -1;

	sprintf(conn->conn_sql,
		"select ri_resvid, "
		"attr_name, "
		"attr_resource, "
		"attr_value, "
		"attr_flags "
		"from pbs.resv_attr "
		"where ri_resvid = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_RESVATTR, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"ri_resvID, "
		"ri_sv_name, "
		"ri_queue, "
		"ri_state, "
		"ri_substate, "
		"ri_type, "
		"ri_stime, "
		"ri_etime, "
		"ri_duration, "
		"ri_tactive, "
		"ri_svrflags, "
		"ri_numattr, "
		"ri_resvTag, "
		"ri_un_type, "
		"ri_fromsock, "
		"ri_fromaddr, "
		"extract(epoch from ri_savetm)::bigint as ri_savetm, "
		"extract(epoch from ri_creattm)::bigint as ri_creattm "
		"from pbs.resv "
		"order by ri_creattm");
	if (pg_prepare_stmt(conn, STMT_FINDRESVS_ORDBY_CREATTM,
		conn->conn_sql, 0) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from pbs.resv where ri_resvid = $1");
	if (pg_prepare_stmt(conn, STMT_DELETE_RESV, conn->conn_sql, 1) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Load resv data from the row into the resv object
 *
 * @param[in]	res - Resultset from a earlier query
 * @param[in]	presv  - resv object to load data into
 * @param[in]	row - The current row to load within the resultset
 *
 */
static void
load_resv(PGresult *res, pbs_db_resv_info_t *presv, int row)
{
	strcpy(presv->ri_resvid, PQgetvalue(res, row,
		PQfnumber(res, "ri_resvID")));
	strcpy(presv->ri_queue, PQgetvalue(res, row,
		PQfnumber(res, "ri_queue")));
	presv->ri_state = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_state")), NULL, 10);
	presv->ri_substate = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_substate")), NULL, 10);
	presv->ri_type = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_type")), NULL, 10);
	presv->ri_stime = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ri_stime")), NULL, 10);
	presv->ri_etime = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ri_etime")), NULL, 10);
	presv->ri_duration = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ri_duration")), NULL, 10);
	presv->ri_tactive = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_tactive")), NULL, 10);
	presv->ri_svrflags = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_svrflags")), NULL, 10);
	presv->ri_numattr = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_numattr")), NULL, 10);
	presv->ri_resvTag = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_resvTag")), NULL, 10);
	presv->ri_un_type = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_un_type")), NULL, 10);
	presv->ri_fromsock = strtol(PQgetvalue(res, row,
		PQfnumber(res, "ri_fromsock")), NULL, 10);
	presv->ri_fromaddr = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ri_fromaddr")), NULL, 10);
	strcpy(presv->ri_sv_name, PQgetvalue(res, row,
		PQfnumber(res, "ri_sv_name")));
	presv->ri_savetm = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ri_savetm")), NULL, 10);
	presv->ri_creattm = strtoll(PQgetvalue(res, row,
		PQfnumber(res, "ri_creattm")), NULL, 10);
}

/**
 * @brief
 *	Insert resv data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of resv to be inserted
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_insert_resv(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;

	LOAD_STR(conn, presv->ri_resvid, 0);
	LOAD_STR(conn, presv->ri_sv_name, 1);
	LOAD_STR(conn, presv->ri_queue, 2);
	LOAD_INTEGER(conn, presv->ri_state, 3);
	LOAD_INTEGER(conn, presv->ri_substate, 4);
	LOAD_INTEGER(conn, presv->ri_type, 5);
	LOAD_BIGINT(conn, presv->ri_stime, 6);
	LOAD_BIGINT(conn, presv->ri_etime, 7);
	LOAD_BIGINT(conn, presv->ri_duration, 8);
	LOAD_INTEGER(conn, presv->ri_tactive, 9);
	LOAD_INTEGER(conn, presv->ri_svrflags, 10);
	LOAD_INTEGER(conn, presv->ri_numattr, 11);
	LOAD_INTEGER(conn, presv->ri_resvTag, 12);
	LOAD_INTEGER(conn, presv->ri_un_type, 13);
	LOAD_INTEGER(conn, presv->ri_fromsock, 14);
	LOAD_BIGINT(conn, presv->ri_fromaddr, 15);

	if (pg_db_cmd(conn, STMT_INSERT_RESV, 16) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Update resv data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of resv to be updated
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows updated
 *
 */
int
pg_db_update_resv(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;

	LOAD_STR(conn, presv->ri_resvid, 0);
	LOAD_STR(conn, presv->ri_sv_name, 1);
	LOAD_STR(conn, presv->ri_queue, 2);
	LOAD_INTEGER(conn, presv->ri_state, 3);
	LOAD_INTEGER(conn, presv->ri_substate, 4);
	LOAD_INTEGER(conn, presv->ri_type, 5);
	LOAD_BIGINT(conn, presv->ri_stime, 6);
	LOAD_BIGINT(conn, presv->ri_etime, 7);
	LOAD_BIGINT(conn, presv->ri_duration, 8);
	LOAD_INTEGER(conn, presv->ri_tactive, 9);
	LOAD_INTEGER(conn, presv->ri_svrflags, 10);
	LOAD_INTEGER(conn, presv->ri_numattr, 11);
	LOAD_INTEGER(conn, presv->ri_resvTag, 12);
	LOAD_INTEGER(conn, presv->ri_un_type, 13);
	LOAD_INTEGER(conn, presv->ri_fromsock, 14);
	LOAD_BIGINT(conn, presv->ri_fromaddr, 15);

	return (pg_db_cmd(conn, STMT_UPDATE_RESV, 16));
}

/**
 * @brief
 *	Load resv data from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Load resv information into this object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows loaded
 *
 */
int
pg_db_load_resv(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;
	PGresult *res;
	int rc;

	LOAD_STR(conn, presv->ri_resvid, 0);

	if ((rc = pg_db_query(conn, STMT_SELECT_RESV, 1, &res)) != 0)
		return rc;

	/* get the other fields */
	load_resv(res, presv, 0);

	PQclear(res);
	return 0;
}

/**
 * @brief
 *	Find resv
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state variable updated by this query
 * @param[in]	obj  - Information of resv to be found
 * @param[in]	opts - Any other options (like flags, timestamp)
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success, but no rows found
 *
 */
int
pg_db_find_resv(pbs_db_conn_t *conn, void *st, pbs_db_obj_info_t *obj,
	pbs_db_query_options_t *opts)
{
	PGresult *res;
	int rc;
	int params;

	pg_query_state_t *state = (pg_query_state_t *) st;

	if (!state)
		return -1;

	params=0;
	strcpy(conn->conn_sql, STMT_FINDRESVS_ORDBY_CREATTM);

	if ((rc = pg_db_query(conn, conn->conn_sql, params, &res)) != 0)
		return rc;

	state->row = 0;
	state->res = res;
	state->count = PQntuples(res);
	return 0;
}

/**
 * @brief
 *	Get the next resv from the cursor
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state
 * @param[in]	obj  - Resv information is loaded into this object
 *
 * @return      Error code
 *		(Even though this returns only 0 now, keeping it as int
 *			to support future change to return a failure)
 * @retval	 0 - Success
 *
 */
int
pg_db_next_resv(pbs_db_conn_t* conn, void* st, pbs_db_obj_info_t* obj)
{
	pg_query_state_t *state = (pg_query_state_t *) st;

	load_resv(state->res, obj->pbs_db_un.pbs_db_resv, state->row);
	return 0;
}

/**
 * @brief
 *	Delete the resv from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Resv information
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows deleted
 *
 */
int
pg_db_delete_resv(pbs_db_conn_t *conn, pbs_db_obj_info_t* obj)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;
	LOAD_STR(conn, presv->ri_resvid, 0);
	return (pg_db_cmd(conn, STMT_DELETE_RESV, 1));
}
