/*
 * Copyright (C) 1994-2016 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *  
 * This file is part of the PBS Professional ("PBS Pro") software.
 * 
 * Open Source License Information:
 *  
 * PBS Pro is free software. You can redistribute it and/or modify it under the
 * terms of the GNU Affero General Public License as published by the Free 
 * Software Foundation, either version 3 of the License, or (at your option) any 
 * later version.
 *  
 * PBS Pro is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License along 
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Commercial License Information: 
 * 
 * The PBS Pro software is licensed under the terms of the GNU Affero General 
 * Public License agreement ("AGPL"), except where a separate commercial license 
 * agreement for PBS Pro version 14 or later has been executed in writing with Altair.
 *  
 * Altair’s dual-license business model allows companies, individuals, and 
 * organizations to create proprietary derivative works of PBS Pro and distribute 
 * them - whether embedded or bundled with other software - under a commercial 
 * license agreement.
 * 
 * Use of Altair’s trademarks, including but not limited to "PBS™", 
 * "PBS Professional®", and "PBS Pro™" and Altair’s logos is subject to Altair's 
 * trademark licensing policies.
 *
 */


/**
 * @file    db_postgres_sched.c
 *
 * @brief
 *      Implementation of the scheduler data access functions for postgres
 */
#include <pbs_config.h>   /* the master config generated by configure */
#include "pbs_db.h"
#include "db_postgres.h"

/**
 * @brief
 *	Prepare all the scheduler related sqls. Typically called after connect
 *	and before any other sql exeuction
 *
 * @param[in]	conn - Database connection handle
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_prepare_sched_sqls(pbs_db_conn_t *conn)
{
	sprintf(conn->conn_sql, "insert into "
		"pbs.scheduler( "
		"sched_name, "
		"sched_sv_name, "
		"sched_savetm, "
		"sched_creattm "
		") "
		"values ($1, $2, localtimestamp, localtimestamp)");
	if (pg_prepare_stmt(conn, STMT_INSERT_SCHED, conn->conn_sql, 2) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.scheduler set "
		"sched_sv_name = $2, "
		"sched_savetm = localtimestamp "
		"where sched_name = $1");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SCHED, conn->conn_sql, 2) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"sched_name, "
		"sched_sv_name, "
		"extract(epoch from sched_savetm)::bigint as sched_savetm, "
		"extract(epoch from sched_creattm)::bigint as sched_creattm "
		"from "
		"pbs.scheduler "
		"where sched_name = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_SCHED, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "insert into "
		"pbs.scheduler_attr "
		"(sched_name, "
		"attr_name, "
		"attr_resource, "
		"attr_value, "
		"attr_flags) "
		"values "
		"($1, $2, $3, $4, $5)");
	if (pg_prepare_stmt(conn, STMT_INSERT_SCHEDATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.scheduler_attr set "
		"attr_resource = $3, "
		"attr_value = $4, "
		"attr_flags = $5 "
		"where sched_name = $1 "
		"and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SCHEDATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.scheduler_attr set "
		"attr_value = $4, "
		"attr_flags = $5 "
		"where sched_name = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SCHEDATTR_RESC,
		conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from "
		"pbs.scheduler_attr "
		"where sched_name = $1 "
		"and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_DELETE_SCHEDATTR, conn->conn_sql, 2) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from "
		"pbs.scheduler_attr "
		"where sched_name = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_DELETE_SCHEDATTR_RESC,
		conn->conn_sql, 3) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"attr_name, "
		"attr_resource, "
		"attr_value, "
		"attr_flags "
		"from "
		"pbs.scheduler_attr "
		"where sched_name = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_SCHEDATTR, conn->conn_sql, 1) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Insert scheduler data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of scheduler to be inserted
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_insert_sched(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_sched_info_t *psch = obj->pbs_db_un.pbs_db_sched;

	LOAD_STR(conn, psch->sched_name, 0);
	LOAD_STR(conn, psch->sched_sv_name, 1);

	if (pg_db_cmd(conn, STMT_INSERT_SCHED, 2) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Update scheduler data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of scheduler to be updated
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows updated
 */
int
pg_db_update_sched(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_sched_info_t *psch = obj->pbs_db_un.pbs_db_sched;

	LOAD_STR(conn, psch->sched_name, 0);
	LOAD_STR(conn, psch->sched_sv_name, 1);

	return (pg_db_cmd(conn, STMT_UPDATE_SCHED, 2));
}

/**
 * @brief
 *	Load scheduler data from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Load scheduler information into this object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows loaded
 *
 */
int
pg_db_load_sched(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	PGresult *res;
	int rc;
	pbs_db_sched_info_t *psch = obj->pbs_db_un.pbs_db_sched;

	LOAD_STR(conn, psch->sched_name, 0);
	if ((rc = pg_db_query(conn, STMT_SELECT_SCHED,  1, &res)) != 0)
		return rc;

	/* get the other fields */
	strcpy(psch->sched_name, PQgetvalue(res, 0,
		PQfnumber(res, "sched_name")));
	strcpy(psch->sched_sv_name, PQgetvalue(res, 0,
		PQfnumber(res, "sched_sv_name")));
	psch->sched_savetm = strtoll(PQgetvalue(res, 0,
		PQfnumber(res, "sched_savetm")), NULL, 10);
	psch->sched_creattm = strtoll(PQgetvalue(res, 0,
		PQfnumber(res, "sched_creattm")), NULL, 10);

	PQclear(res);
	return 0;
}
