/*
 * Copyright (C) 1994-2016 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *  
 * This file is part of the PBS Professional ("PBS Pro") software.
 * 
 * Open Source License Information:
 *  
 * PBS Pro is free software. You can redistribute it and/or modify it under the
 * terms of the GNU Affero General Public License as published by the Free 
 * Software Foundation, either version 3 of the License, or (at your option) any 
 * later version.
 *  
 * PBS Pro is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License along 
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Commercial License Information: 
 * 
 * The PBS Pro software is licensed under the terms of the GNU Affero General 
 * Public License agreement ("AGPL"), except where a separate commercial license 
 * agreement for PBS Pro version 14 or later has been executed in writing with Altair.
 *  
 * Altair’s dual-license business model allows companies, individuals, and 
 * organizations to create proprietary derivative works of PBS Pro and distribute 
 * them - whether embedded or bundled with other software - under a commercial 
 * license agreement.
 * 
 * Use of Altair’s trademarks, including but not limited to "PBS™", 
 * "PBS Professional®", and "PBS Pro™" and Altair’s logos is subject to Altair's 
 * trademark licensing policies.
 *
 */


/**
 * @file    db_postgres_svr.c
 *
 * @brief
 *      Implementation of the svr data access functions for postgres
 */

#include <pbs_config.h>   /* the master config generated by configure */
#include "pbs_db.h"
#include <errno.h>
#include "db_postgres.h"

/**
 * @brief
 *	Prepare all the server related sqls. Typically called after connect
 *	and before any other sql exeuction
 *
 * @param[in]	conn - Database connection handle
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_prepare_svr_sqls(pbs_db_conn_t *conn)
{
	sprintf(conn->conn_sql, "insert into pbs.server( "
		"sv_name, "
		"sv_hostname, "
		"sv_numjobs, "
		"sv_numque, "
		"sv_jobidnumber, "
		"sv_svraddr, "
		"sv_svrport, "
		"sv_savetm, "
		"sv_creattm "
		") "
		"values "
		"($1, $2, $3, $4, $5, $6, $7, localtimestamp, localtimestamp)");
	if (pg_prepare_stmt(conn, STMT_INSERT_SVR, conn->conn_sql, 7) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.server set "
		"sv_hostname = $2, "
		"sv_numjobs = $3, "
		"sv_numque = $4, "
		"sv_jobidnumber = $5, "
		"sv_svraddr = $6, "
		"sv_svrport = $7, "
		"sv_savetm = localtimestamp "
		"where sv_name = $1");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SVR, conn->conn_sql, 7) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"sv_name, "
		"sv_hostname, "
		"sv_numjobs, "
		"sv_numque, "
		"sv_jobidnumber, "
		"extract(epoch from sv_savetm)::bigint as sv_savetm, "
		"extract(epoch from sv_creattm)::bigint as sv_creattm "
		"from "
		"pbs.server where sv_name = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_SVR, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "insert into "
		"pbs.server_attr "
		"(sv_name, "
		"attr_name, "
		"attr_resource, "
		"attr_value, "
		"attr_flags) "
		"values "
		"($1, $2, $3, $4, $5)");
	if (pg_prepare_stmt(conn, STMT_INSERT_SVRATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.server_attr set "
		"attr_resource = $3, "
		"attr_value = $4, "
		"attr_flags = $5 "
		"where sv_name = $1 "
		"and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SVRATTR, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "update pbs.server_attr set "
		"attr_value = $4, "
		"attr_flags = $5 "
		"where sv_name = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_UPDATE_SVRATTR_RESC, conn->conn_sql, 5) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from "
		"pbs.server_attr "
		"where sv_name = $1");
	if (pg_prepare_stmt(conn, STMT_DELETE_SVRATTR_ALL, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from "
		"pbs.server_attr "
		"where sv_name = $1 "
		"and attr_name = $2");
	if (pg_prepare_stmt(conn, STMT_DELETE_SVRATTR, conn->conn_sql, 2) != 0)
		return -1;

	sprintf(conn->conn_sql, "delete from "
		"pbs.server_attr "
		"where sv_name = $1 "
		"and attr_name = $2 "
		"and attr_resource = $3");
	if (pg_prepare_stmt(conn, STMT_DELETE_SVRATTR_RESC, conn->conn_sql, 3) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"attr_name, "
		"attr_resource, "
		"attr_value, "
		"attr_flags "
		"from pbs.server_attr "
		"where sv_name = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_SVRATTR, conn->conn_sql, 1) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"pbs_schema_version "
		"from "
		"pbs.info");
	if (pg_prepare_stmt(conn, STMT_SELECT_DBVER, conn->conn_sql, 0) != 0)
		return -1;

	sprintf(conn->conn_sql, "select nextval('pbs.svr_id_seq')");
	if (pg_prepare_stmt(conn, STMT_SELECT_NEXT_SEQID, conn->conn_sql, 0) != 0)
		return -1;

	sprintf(conn->conn_sql, "select "
		"sv_name "
		"from "
		"pbs.server where sv_hostname = $1");
	if (pg_prepare_stmt(conn, STMT_SELECT_SVRID, conn->conn_sql, 1) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Truncate all data from ALL tables from the database
 *
 * @param[in]	conn - The database connection handle
 *
 * @return      Error code
 * @retval	-1 - Failure
 *		 0 - Success
 *
 */
int
pbs_db_truncate_all(pbs_db_conn_t *conn)
{
	sprintf(conn->conn_sql, "truncate table 	"
		"pbs.scheduler_attr, "
		"pbs.scheduler, "
		"pbs.node_attr, "
		"pbs.node, "
		"pbs.queue_attr, "
		"pbs.queue, "
		"pbs.resv_attr, "
		"pbs.resv, "
		"pbs.subjob_track, "
		"pbs.job_scr, "
		"pbs.job_attr, "
		"pbs.job, "
		"pbs.server_attr, "
		"pbs.server");

	if (pbs_db_execute_str(conn, conn->conn_sql) == -1)
		return -1;

	return 0;
}

/**
 * @brief
 *	Insert server data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of server to be inserted
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pg_db_insert_svr(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_svr_info_t *ps = obj->pbs_db_un.pbs_db_svr;

	LOAD_STR(conn, ps->sv_name, 0);
	LOAD_STR(conn, ps->sv_hostname, 1);
	LOAD_INTEGER(conn, ps->sv_numjobs, 2);
	LOAD_INTEGER(conn, ps->sv_numque, 3);
	LOAD_INTEGER(conn, ps->sv_jobidnumber, 4);
	LOAD_BIGINT(conn, ps->sv_svraddr, 5);
	LOAD_INTEGER(conn, ps->sv_svrport, 6);

	if (pg_db_cmd(conn, STMT_INSERT_SVR, 7) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Update server data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of server to be updated
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows updated
 *
 */
int
pg_db_update_svr(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_svr_info_t *ps = obj->pbs_db_un.pbs_db_svr;

	LOAD_STR(conn, ps->sv_name, 0);
	LOAD_STR(conn, ps->sv_hostname, 1);
	LOAD_INTEGER(conn, ps->sv_numjobs, 2);
	LOAD_INTEGER(conn, ps->sv_numque, 3);
	LOAD_INTEGER(conn, ps->sv_jobidnumber, 4);
	LOAD_BIGINT(conn, ps->sv_svraddr, 5);
	LOAD_INTEGER(conn, ps->sv_svrport, 6);

	return (pg_db_cmd(conn, STMT_UPDATE_SVR, 7));
}

/**
 * @brief
 *	Load server data from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Load server information into this object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows loaded
 *
 */
int
pg_db_load_svr(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	PGresult *res;
	int rc;
	pbs_db_svr_info_t *ps = obj->pbs_db_un.pbs_db_svr;

	LOAD_STR(conn, ps->sv_name, 0);
	if ((rc = pg_db_query(conn, STMT_SELECT_SVR,  1, &res)) != 0)
		return rc;

	/* get the other fields */
	strcpy(ps->sv_name, PQgetvalue(res, 0,
		PQfnumber(res, "sv_name")));
	strcpy(ps->sv_hostname, PQgetvalue(res, 0,
		PQfnumber(res, "sv_hostname")));
	ps->sv_numjobs = strtol(PQgetvalue(res, 0,
		PQfnumber(res, "sv_numjobs")), NULL, 10);
	ps->sv_numque = strtol(PQgetvalue(res, 0,
		PQfnumber(res, "sv_numque")), NULL, 10);
	ps->sv_jobidnumber = strtol(PQgetvalue(res, 0,
		PQfnumber(res, "sv_jobidnumber")), NULL, 10);
	ps->sv_savetm = strtoll(PQgetvalue(res, 0,
		PQfnumber(res, "sv_savetm")), NULL, 10);
	ps->sv_creattm = strtoll(PQgetvalue(res, 0,
		PQfnumber(res, "sv_creattm")), NULL, 10);

	PQclear(res);
	return 0;
}

/**
 * @brief
 *	Retrieve the Datastore schema version (maj, min)
 *
 * @param[out]   db_maj_ver - return the major schema version
 * @param[out]   db_min_ver - return the minor schema version
 *
 * @return     Error code
 * @retval     -1 - Failure
 * @retval     0  - Success
 *
 */
int
pbs_db_get_schema_version(pbs_db_conn_t *conn, int *db_maj_ver, int *db_min_ver)
{
	PGresult *res;
	int rc;
	char ver_str[MAX_SCHEMA_VERSION_LEN+1];
	char *token;

	if ((rc = pg_db_query(conn, STMT_SELECT_DBVER, 0, &res)) != 0)
		return rc;

	memset(ver_str, 0, sizeof(ver_str));
	strncpy(ver_str, PQgetvalue(res, 0, PQfnumber(res, "pbs_schema_version")), MAX_SCHEMA_VERSION_LEN+1);
	if (ver_str[MAX_SCHEMA_VERSION_LEN] != '\0')
		return -1;

	token = strtok(ver_str, ".");
	if (token)
		*db_maj_ver = atol(token);
	else
		return -1;

	token = strtok(NULL, ".");
	if (token)
		*db_min_ver = atol(token);
	else
		return -1;

	return 0;
}

/**
 * @brief
 *	Get the svrid sequence number from the database
 *
 * The svr hostname is associated with a sever_id (sv_name) column. This column
 * serves as the id for the server, used for all subsequent queries. These "id"
 * values are created from a sequence generator within the PBS schema called
 * "pbs.svr_id_seq", whenever a new server database is created (typically at new
 * installation).
 * This function reads the "next" value from the sequence "pbs.svr_id_seq" and
 * returns that number (in ascii format) to be used as the server-id for the
 * new server database being created.
 *
 * @see chk_and_update_db_svrhost
 *
 * @param[in]	conn - The database connection handle
 *
 * @return      Current sequence number in ascii format (to be freed by caller)
 * @retval	-NULL  - Failure
 *		-!NULL - Success
 *
 */
char*
pbs_db_get_unique_svrid(pbs_db_conn_t *conn)
{
	PGresult *res;
	char *svrid;

	if (pg_db_query(conn, STMT_SELECT_NEXT_SEQID, 0, &res) != 0)
		return NULL;

	svrid = strdup(PQgetvalue(res, 0, 0));
	PQclear(res);

	return svrid;
}

/**
 * @brief
 *	Get the svrid corresponding to the given svr hostname
 *
 * The svr hostname is associated with a sever_id (sv_name) column. This column
 * serves as the id for the server, used for all subsequent queries. For a
 * server database that has already been created, an id value (sv_name column)
 * is associated with the server's hostname. This id needs to be retrieved by
 * pbs_server to be used in subsequent queries for this server row.
 *
 * This function loads the server_id (sv_name column value) corresponding to the
 * given hostname (sv_hostname column).
 *
 * @see chk_and_update_db_svrhost
 *
 * @param[in]	conn - The database connection handle
 * @param[in]	hostname - The svr hostname
 *
 * @return      The database svrid (to be freed by caller)
 * @retval	-NULL  - Failure
 *		-!NULL - Success
 *
 */
char*
pbs_db_get_svr_id(pbs_db_conn_t *conn, char *hostname)
{
	PGresult *res;
	char *svrid;

	LOAD_STR(conn, hostname, 0);
	if (pg_db_query(conn, STMT_SELECT_SVRID, 1, &res) != 0)
		return NULL;

	svrid = strdup(PQgetvalue(res, 0, 0));
	PQclear(res);

	return svrid;
}

/**
 * @brief
 *	Delete all the server attributes from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - server information
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success but no rows deleted
 *
 */
int
pg_db_delete_svrattr(pbs_db_conn_t *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_svr_info_t *ps = obj->pbs_db_un.pbs_db_svr;
	LOAD_STR(conn, ps->sv_name, 0);

	return (pg_db_cmd(conn, STMT_DELETE_SVRATTR_ALL, 1));
}
