/*
 * Copyright (C) 1994-2016 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *  
 * This file is part of the PBS Professional ("PBS Pro") software.
 * 
 * Open Source License Information:
 *  
 * PBS Pro is free software. You can redistribute it and/or modify it under the
 * terms of the GNU Affero General Public License as published by the Free 
 * Software Foundation, either version 3 of the License, or (at your option) any 
 * later version.
 *  
 * PBS Pro is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License along 
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * Commercial License Information: 
 * 
 * The PBS Pro software is licensed under the terms of the GNU Affero General 
 * Public License agreement ("AGPL"), except where a separate commercial license 
 * agreement for PBS Pro version 14 or later has been executed in writing with Altair.
 *  
 * Altair’s dual-license business model allows companies, individuals, and 
 * organizations to create proprietary derivative works of PBS Pro and distribute 
 * them - whether embedded or bundled with other software - under a commercial 
 * license agreement.
 * 
 * Use of Altair’s trademarks, including but not limited to "PBS™", 
 * "PBS Professional®", and "PBS Pro™" and Altair’s logos is subject to Altair's 
 * trademark licensing policies.
 *
 */
/**
 * @file	req_getcred.c
 *
 * This file contains function relating to the PBS credential system,
 * it includes the major functions:
 *   req_authenResvPort - Authenticate a user connection based on pbs_iff  (new)
 *   req_connect    - validate the credential in a Connection Request (old)
 */
#include <pbs_config.h>   /* the master config generated by configure */

#include <sys/types.h>
#include <string.h>
#include "libpbs.h"
#include "server_limits.h"
#include "list_link.h"
#include "attribute.h"
#include "credential.h"
#include "net_connect.h"
#include "batch_request.h"


/* External Global Data Items Referenced */

extern time_t time_now;

extern struct connection *svr_conn;
extern int max_connection;

/* Global Data Home in this file */

/**
 * @brief
 * 		req_connect - process a Connection Request
 * 		Almost does nothing.
 *
 * @param[in]	preq	- Connection Request
 */

void
req_connect(struct batch_request *preq)
{
	int  conn_idx = connection_find_actual_index(preq->rq_conn);

	if (conn_idx == -1) {
		req_reject(PBSE_SYSTEM, 0, preq);
		return;
	}

	if ( (preq->rq_extend != NULL) && \
		(strcmp(preq->rq_extend, QSUB_DAEMON) == 0) ) {
		svr_conn[conn_idx].cn_authen |= PBS_NET_CONN_FROM_QSUB_DAEMON;
	}

	if ((svr_conn[conn_idx].cn_authen &
		(PBS_NET_CONN_AUTHENTICATED|PBS_NET_CONN_FROM_PRIVIL))==0) {
		reply_ack(preq);
	} else
		req_reject(PBSE_BADCRED, 0, preq);
}

/** 
 * @brief     
 *		req_authenResvPort - Authenticate a user connection based on the (new)
 *		pbs_iff information.  Pbs_iff will contact the server on a privileged
 *		port and identify the user who has made an existing, but yet unused,
 *		non-privileged connection.  This connection is marked as authenticated.
 *
 * @param[in]	preq	- Connection Request
 */

void
req_authenResvPort(struct batch_request *preq)
{
	int		s;
	pbs_net_t	req_addr;
	int		conn_idx;
	uint		authrequest_port = preq->rq_ind.rq_authen_resvport.rq_port;

	conn_idx = connection_find_actual_index(preq->rq_conn);
	if (conn_idx == -1) {
		req_reject(PBSE_SYSTEM, 0, preq);
		return;
	}

	req_addr = svr_conn[conn_idx].cn_addr;

	/*
	 * find the socket whose client side is bound to the port named
	 * in the request
	 */

	for (s=0; s<max_connection; ++s) {
		if (authrequest_port == svr_conn[s].cn_port &&
			req_addr == svr_conn[s].cn_addr) {

			if ((svr_conn[s].cn_authen & (PBS_NET_CONN_AUTHENTICATED|PBS_NET_CONN_FROM_PRIVIL))==0) {
				(void)strcpy(svr_conn[s].cn_username,
					preq->rq_user);
				(void)strcpy(svr_conn[s].cn_hostname,
					preq->rq_host);
				/* time stamp just for the record */
				svr_conn[s].cn_timestamp = time_now;
				svr_conn[s].cn_authen |= PBS_NET_CONN_AUTHENTICATED;
			}
			reply_ack(preq);
			return;
		}
	}
	req_reject(PBSE_BADCRED, 0, preq);
}
